/* A Bison parser, made by GNU Bison 3.8.2.  */

/* Bison interface for Yacc-like parsers in C

   Copyright (C) 1984, 1989-1990, 2000-2015, 2018-2021 Free Software Foundation,
   Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

/* As a special exception, you may create a larger work that contains
   part or all of the Bison parser skeleton and distribute that work
   under terms of your choice, so long as that work isn't itself a
   parser generator using the skeleton or a modified version thereof
   as a parser skeleton.  Alternatively, if you modify or redistribute
   the parser skeleton itself, you may (at your option) remove this
   special exception, which will cause the skeleton and the resulting
   Bison output files to be licensed under the GNU General Public
   License without this special exception.

   This special exception was added by the Free Software Foundation in
   version 2.2 of Bison.  */

/* DO NOT RELY ON FEATURES THAT ARE NOT DOCUMENTED in the manual,
   especially those whose name start with YY_ or yy_.  They are
   private implementation details that can be changed or removed.  */

#ifndef YY_YY_CONFIG_GRAM_H_INCLUDED
# define YY_YY_CONFIG_GRAM_H_INCLUDED
/* Debug traces.  */
#ifndef YYDEBUG
# define YYDEBUG 0
#endif
#if YYDEBUG
extern int yydebug;
#endif

/* Token kinds.  */
#ifndef YYTOKENTYPE
# define YYTOKENTYPE
  enum yytokentype
  {
    YYEMPTY = -2,
    YYEOF = 0,                     /* "end of file"  */
    YYerror = 256,                 /* error  */
    YYUNDEF = 257,                 /* "invalid token"  */
    K_READ = 258,                  /* K_READ  */
    K_WRITE = 259,                 /* K_WRITE  */
    K_READ_LO = 260,               /* K_READ_LO  */
    K_READ_HI = 261,               /* K_READ_HI  */
    K_WRITE_LO = 262,              /* K_WRITE_LO  */
    K_WRITE_HI = 263,              /* K_WRITE_HI  */
    K_LOADPAGE_LO = 264,           /* K_LOADPAGE_LO  */
    K_LOADPAGE_HI = 265,           /* K_LOADPAGE_HI  */
    K_LOAD_EXT_ADDR = 266,         /* K_LOAD_EXT_ADDR  */
    K_WRITEPAGE = 267,             /* K_WRITEPAGE  */
    K_CHIP_ERASE = 268,            /* K_CHIP_ERASE  */
    K_PGM_ENABLE = 269,            /* K_PGM_ENABLE  */
    K_MEMORY = 270,                /* K_MEMORY  */
    K_PAGE_SIZE = 271,             /* K_PAGE_SIZE  */
    K_PAGED = 272,                 /* K_PAGED  */
    K_BAUDRATE = 273,              /* K_BAUDRATE  */
    K_BS2 = 274,                   /* K_BS2  */
    K_BUFF = 275,                  /* K_BUFF  */
    K_CHIP_ERASE_DELAY = 276,      /* K_CHIP_ERASE_DELAY  */
    K_CONNTYPE = 277,              /* K_CONNTYPE  */
    K_DEDICATED = 278,             /* K_DEDICATED  */
    K_DEFAULT_BITCLOCK = 279,      /* K_DEFAULT_BITCLOCK  */
    K_DEFAULT_PARALLEL = 280,      /* K_DEFAULT_PARALLEL  */
    K_DEFAULT_PROGRAMMER = 281,    /* K_DEFAULT_PROGRAMMER  */
    K_DEFAULT_SAFEMODE = 282,      /* K_DEFAULT_SAFEMODE  */
    K_DEFAULT_SERIAL = 283,        /* K_DEFAULT_SERIAL  */
    K_DESC = 284,                  /* K_DESC  */
    K_DEVICECODE = 285,            /* K_DEVICECODE  */
    K_STK500_DEVCODE = 286,        /* K_STK500_DEVCODE  */
    K_AVR910_DEVCODE = 287,        /* K_AVR910_DEVCODE  */
    K_EEPROM = 288,                /* K_EEPROM  */
    K_ERRLED = 289,                /* K_ERRLED  */
    K_FLASH = 290,                 /* K_FLASH  */
    K_ID = 291,                    /* K_ID  */
    K_IO = 292,                    /* K_IO  */
    K_LOADPAGE = 293,              /* K_LOADPAGE  */
    K_MAX_WRITE_DELAY = 294,       /* K_MAX_WRITE_DELAY  */
    K_MCU_BASE = 295,              /* K_MCU_BASE  */
    K_MIN_WRITE_DELAY = 296,       /* K_MIN_WRITE_DELAY  */
    K_MISO = 297,                  /* K_MISO  */
    K_MOSI = 298,                  /* K_MOSI  */
    K_NUM_PAGES = 299,             /* K_NUM_PAGES  */
    K_NVM_BASE = 300,              /* K_NVM_BASE  */
    K_OCDREV = 301,                /* K_OCDREV  */
    K_OFFSET = 302,                /* K_OFFSET  */
    K_PAGEL = 303,                 /* K_PAGEL  */
    K_PARALLEL = 304,              /* K_PARALLEL  */
    K_PARENT = 305,                /* K_PARENT  */
    K_PART = 306,                  /* K_PART  */
    K_PGMLED = 307,                /* K_PGMLED  */
    K_PROGRAMMER = 308,            /* K_PROGRAMMER  */
    K_PSEUDO = 309,                /* K_PSEUDO  */
    K_PWROFF_AFTER_WRITE = 310,    /* K_PWROFF_AFTER_WRITE  */
    K_RDYLED = 311,                /* K_RDYLED  */
    K_READBACK_P1 = 312,           /* K_READBACK_P1  */
    K_READBACK_P2 = 313,           /* K_READBACK_P2  */
    K_READMEM = 314,               /* K_READMEM  */
    K_RESET = 315,                 /* K_RESET  */
    K_RETRY_PULSE = 316,           /* K_RETRY_PULSE  */
    K_SERIAL = 317,                /* K_SERIAL  */
    K_SCK = 318,                   /* K_SCK  */
    K_SIGNATURE = 319,             /* K_SIGNATURE  */
    K_SIZE = 320,                  /* K_SIZE  */
    K_USB = 321,                   /* K_USB  */
    K_USBDEV = 322,                /* K_USBDEV  */
    K_USBSN = 323,                 /* K_USBSN  */
    K_USBPID = 324,                /* K_USBPID  */
    K_USBPRODUCT = 325,            /* K_USBPRODUCT  */
    K_USBVENDOR = 326,             /* K_USBVENDOR  */
    K_USBVID = 327,                /* K_USBVID  */
    K_TYPE = 328,                  /* K_TYPE  */
    K_VCC = 329,                   /* K_VCC  */
    K_VFYLED = 330,                /* K_VFYLED  */
    K_NO = 331,                    /* K_NO  */
    K_YES = 332,                   /* K_YES  */
    K_TIMEOUT = 333,               /* K_TIMEOUT  */
    K_STABDELAY = 334,             /* K_STABDELAY  */
    K_CMDEXEDELAY = 335,           /* K_CMDEXEDELAY  */
    K_HVSPCMDEXEDELAY = 336,       /* K_HVSPCMDEXEDELAY  */
    K_SYNCHLOOPS = 337,            /* K_SYNCHLOOPS  */
    K_BYTEDELAY = 338,             /* K_BYTEDELAY  */
    K_POLLVALUE = 339,             /* K_POLLVALUE  */
    K_POLLINDEX = 340,             /* K_POLLINDEX  */
    K_PREDELAY = 341,              /* K_PREDELAY  */
    K_POSTDELAY = 342,             /* K_POSTDELAY  */
    K_POLLMETHOD = 343,            /* K_POLLMETHOD  */
    K_MODE = 344,                  /* K_MODE  */
    K_DELAY = 345,                 /* K_DELAY  */
    K_BLOCKSIZE = 346,             /* K_BLOCKSIZE  */
    K_READSIZE = 347,              /* K_READSIZE  */
    K_HVENTERSTABDELAY = 348,      /* K_HVENTERSTABDELAY  */
    K_PROGMODEDELAY = 349,         /* K_PROGMODEDELAY  */
    K_LATCHCYCLES = 350,           /* K_LATCHCYCLES  */
    K_TOGGLEVTG = 351,             /* K_TOGGLEVTG  */
    K_POWEROFFDELAY = 352,         /* K_POWEROFFDELAY  */
    K_RESETDELAYMS = 353,          /* K_RESETDELAYMS  */
    K_RESETDELAYUS = 354,          /* K_RESETDELAYUS  */
    K_HVLEAVESTABDELAY = 355,      /* K_HVLEAVESTABDELAY  */
    K_RESETDELAY = 356,            /* K_RESETDELAY  */
    K_SYNCHCYCLES = 357,           /* K_SYNCHCYCLES  */
    K_HVCMDEXEDELAY = 358,         /* K_HVCMDEXEDELAY  */
    K_CHIPERASEPULSEWIDTH = 359,   /* K_CHIPERASEPULSEWIDTH  */
    K_CHIPERASEPOLLTIMEOUT = 360,  /* K_CHIPERASEPOLLTIMEOUT  */
    K_CHIPERASETIME = 361,         /* K_CHIPERASETIME  */
    K_PROGRAMFUSEPULSEWIDTH = 362, /* K_PROGRAMFUSEPULSEWIDTH  */
    K_PROGRAMFUSEPOLLTIMEOUT = 363, /* K_PROGRAMFUSEPOLLTIMEOUT  */
    K_PROGRAMLOCKPULSEWIDTH = 364, /* K_PROGRAMLOCKPULSEWIDTH  */
    K_PROGRAMLOCKPOLLTIMEOUT = 365, /* K_PROGRAMLOCKPOLLTIMEOUT  */
    K_PP_CONTROLSTACK = 366,       /* K_PP_CONTROLSTACK  */
    K_HVSP_CONTROLSTACK = 367,     /* K_HVSP_CONTROLSTACK  */
    K_ALLOWFULLPAGEBITSTREAM = 368, /* K_ALLOWFULLPAGEBITSTREAM  */
    K_ENABLEPAGEPROGRAMMING = 369, /* K_ENABLEPAGEPROGRAMMING  */
    K_HAS_JTAG = 370,              /* K_HAS_JTAG  */
    K_HAS_DW = 371,                /* K_HAS_DW  */
    K_HAS_PDI = 372,               /* K_HAS_PDI  */
    K_HAS_TPI = 373,               /* K_HAS_TPI  */
    K_IDR = 374,                   /* K_IDR  */
    K_IS_AT90S1200 = 375,          /* K_IS_AT90S1200  */
    K_IS_AVR32 = 376,              /* K_IS_AVR32  */
    K_RAMPZ = 377,                 /* K_RAMPZ  */
    K_SPMCR = 378,                 /* K_SPMCR  */
    K_EECR = 379,                  /* K_EECR  */
    K_FLASH_INSTR = 380,           /* K_FLASH_INSTR  */
    K_EEPROM_INSTR = 381,          /* K_EEPROM_INSTR  */
    TKN_COMMA = 382,               /* TKN_COMMA  */
    TKN_EQUAL = 383,               /* TKN_EQUAL  */
    TKN_SEMI = 384,                /* TKN_SEMI  */
    TKN_TILDE = 385,               /* TKN_TILDE  */
    TKN_LEFT_PAREN = 386,          /* TKN_LEFT_PAREN  */
    TKN_RIGHT_PAREN = 387,         /* TKN_RIGHT_PAREN  */
    TKN_NUMBER = 388,              /* TKN_NUMBER  */
    TKN_NUMBER_REAL = 389,         /* TKN_NUMBER_REAL  */
    TKN_STRING = 390               /* TKN_STRING  */
  };
  typedef enum yytokentype yytoken_kind_t;
#endif
/* Token kinds.  */
#define YYEMPTY -2
#define YYEOF 0
#define YYerror 256
#define YYUNDEF 257
#define K_READ 258
#define K_WRITE 259
#define K_READ_LO 260
#define K_READ_HI 261
#define K_WRITE_LO 262
#define K_WRITE_HI 263
#define K_LOADPAGE_LO 264
#define K_LOADPAGE_HI 265
#define K_LOAD_EXT_ADDR 266
#define K_WRITEPAGE 267
#define K_CHIP_ERASE 268
#define K_PGM_ENABLE 269
#define K_MEMORY 270
#define K_PAGE_SIZE 271
#define K_PAGED 272
#define K_BAUDRATE 273
#define K_BS2 274
#define K_BUFF 275
#define K_CHIP_ERASE_DELAY 276
#define K_CONNTYPE 277
#define K_DEDICATED 278
#define K_DEFAULT_BITCLOCK 279
#define K_DEFAULT_PARALLEL 280
#define K_DEFAULT_PROGRAMMER 281
#define K_DEFAULT_SAFEMODE 282
#define K_DEFAULT_SERIAL 283
#define K_DESC 284
#define K_DEVICECODE 285
#define K_STK500_DEVCODE 286
#define K_AVR910_DEVCODE 287
#define K_EEPROM 288
#define K_ERRLED 289
#define K_FLASH 290
#define K_ID 291
#define K_IO 292
#define K_LOADPAGE 293
#define K_MAX_WRITE_DELAY 294
#define K_MCU_BASE 295
#define K_MIN_WRITE_DELAY 296
#define K_MISO 297
#define K_MOSI 298
#define K_NUM_PAGES 299
#define K_NVM_BASE 300
#define K_OCDREV 301
#define K_OFFSET 302
#define K_PAGEL 303
#define K_PARALLEL 304
#define K_PARENT 305
#define K_PART 306
#define K_PGMLED 307
#define K_PROGRAMMER 308
#define K_PSEUDO 309
#define K_PWROFF_AFTER_WRITE 310
#define K_RDYLED 311
#define K_READBACK_P1 312
#define K_READBACK_P2 313
#define K_READMEM 314
#define K_RESET 315
#define K_RETRY_PULSE 316
#define K_SERIAL 317
#define K_SCK 318
#define K_SIGNATURE 319
#define K_SIZE 320
#define K_USB 321
#define K_USBDEV 322
#define K_USBSN 323
#define K_USBPID 324
#define K_USBPRODUCT 325
#define K_USBVENDOR 326
#define K_USBVID 327
#define K_TYPE 328
#define K_VCC 329
#define K_VFYLED 330
#define K_NO 331
#define K_YES 332
#define K_TIMEOUT 333
#define K_STABDELAY 334
#define K_CMDEXEDELAY 335
#define K_HVSPCMDEXEDELAY 336
#define K_SYNCHLOOPS 337
#define K_BYTEDELAY 338
#define K_POLLVALUE 339
#define K_POLLINDEX 340
#define K_PREDELAY 341
#define K_POSTDELAY 342
#define K_POLLMETHOD 343
#define K_MODE 344
#define K_DELAY 345
#define K_BLOCKSIZE 346
#define K_READSIZE 347
#define K_HVENTERSTABDELAY 348
#define K_PROGMODEDELAY 349
#define K_LATCHCYCLES 350
#define K_TOGGLEVTG 351
#define K_POWEROFFDELAY 352
#define K_RESETDELAYMS 353
#define K_RESETDELAYUS 354
#define K_HVLEAVESTABDELAY 355
#define K_RESETDELAY 356
#define K_SYNCHCYCLES 357
#define K_HVCMDEXEDELAY 358
#define K_CHIPERASEPULSEWIDTH 359
#define K_CHIPERASEPOLLTIMEOUT 360
#define K_CHIPERASETIME 361
#define K_PROGRAMFUSEPULSEWIDTH 362
#define K_PROGRAMFUSEPOLLTIMEOUT 363
#define K_PROGRAMLOCKPULSEWIDTH 364
#define K_PROGRAMLOCKPOLLTIMEOUT 365
#define K_PP_CONTROLSTACK 366
#define K_HVSP_CONTROLSTACK 367
#define K_ALLOWFULLPAGEBITSTREAM 368
#define K_ENABLEPAGEPROGRAMMING 369
#define K_HAS_JTAG 370
#define K_HAS_DW 371
#define K_HAS_PDI 372
#define K_HAS_TPI 373
#define K_IDR 374
#define K_IS_AT90S1200 375
#define K_IS_AVR32 376
#define K_RAMPZ 377
#define K_SPMCR 378
#define K_EECR 379
#define K_FLASH_INSTR 380
#define K_EEPROM_INSTR 381
#define TKN_COMMA 382
#define TKN_EQUAL 383
#define TKN_SEMI 384
#define TKN_TILDE 385
#define TKN_LEFT_PAREN 386
#define TKN_RIGHT_PAREN 387
#define TKN_NUMBER 388
#define TKN_NUMBER_REAL 389
#define TKN_STRING 390

/* Value type.  */
#if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
typedef int YYSTYPE;
# define YYSTYPE_IS_TRIVIAL 1
# define YYSTYPE_IS_DECLARED 1
#endif


extern YYSTYPE yylval;


int yyparse (void);


#endif /* !YY_YY_CONFIG_GRAM_H_INCLUDED  */
